"""MySQL target sink class, which handles writing streams."""

from __future__ import annotations

import json
import logging
import re
import typing as t
from datetime import datetime, timezone
import sys
from typing import Any, Dict, Iterable, List, Optional, cast

import sqlalchemy
from c8connector import ensure_datetime
from singer_sdk.connectors import SQLConnector
from singer_sdk.helpers._conformers import replace_leading_digit
from singer_sdk.helpers._typing import get_datelike_property_type
from singer_sdk.sinks import SQLSink
from sqlalchemy import Column
from sqlalchemy.dialects import mysql
from sqlalchemy.engine import URL, Engine
from sqlalchemy.schema import PrimaryKeyConstraint

from macrometa_target_mysql.constants import (
    export_errors,
    export_lag,
    fabric_label,
    region_label,
    tenant_label,
    workflow_label,
)

if t.TYPE_CHECKING:
    from sqlalchemy.engine.reflection import Inspector


class MySQLConnector(SQLConnector):
    """The connector for MySQL.

    This class handles all DDL and type conversions.
    """

    allow_column_add: bool = True  # Whether ADD COLUMN is supported.
    allow_column_rename: bool = True  # Whether RENAME COLUMN is supported.
    allow_column_alter: bool = False  # Whether altering column types is supported.
    allow_merge_upsert: bool = False  # Whether MERGE UPSERT is supported.
    allow_temp_tables: bool = True  # Whether temp tables are supported.
    table_name_pattern: str = (
        "${TABLE_NAME}"  # The pattern to use for temp table names.
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger.setLevel(logging.DEBUG)

        if "allow_column_alter" in super().config:
            self.allow_column_alter = super().config.get("allow_column_alter")

    def get_sqlalchemy_url(self, config: dict) -> URL:
        """Generates a SQLAlchemy URL for MySQL.

        Args:
            config: The configuration for the connector.
        """

        if config.get("sqlalchemy_url"):
            return config["sqlalchemy_url"]

        return sqlalchemy.engine.url.URL.create(
            drivername="mysql",
            username=config["username"],
            password=config["password"],
            host=config["host"],
            port=config["port"],
            database=config["database"],
        )

    def get_fully_qualified_name(
        self,
        table_name: str | None = None,
        schema_name: str | None = None,
        db_name: str | None = None,
        delimiter: str = ".",
    ) -> str:
        """Concatenates a fully qualified name from the parts.

        Args:
            table_name: The name of the table.
            schema_name: The name of the schema. Defaults to None.
            db_name: The name of the database. Defaults to None.
            delimiter: Generally: '.' for SQL names and '-' for Singer names.

        Raises:
            ValueError: If all 3 name parts not supplied.

        Returns:
            The fully qualified name as a string.
        """
        target_table = self.config.get("target_table")
        if not target_table:
            target_table = table_name

        parts = []
        if db_name:
            parts.append(db_name)
        if schema_name:
            parts.append(schema_name)
        if table_name:
            parts.append(target_table)

        if not parts:
            raise ValueError(
                "Could not generate fully qualified name: "
                + ":".join(
                    [
                        db_name or "(unknown-db)",
                        schema_name or "(unknown-schema)",
                        table_name or "(unknown-table-name)",
                    ],
                ),
            )

        return delimiter.join(parts)

    def get_object_names(
        self,
        engine: Engine,  # noqa: ARG002
        inspected: Inspector,
        schema_name: str,
    ) -> list[tuple[str, bool]]:
        """Return a list of syncable objects.
        Args:
            engine: SQLAlchemy engine
            inspected: SQLAlchemy inspector instance for engine
            schema_name: Schema name to inspect

        Returns:
            List of tuples (<table_or_view_name>, <is_view>)
        """
        # Get list of tables and views
        table_names = inspected.get_table_names(schema=schema_name)
        try:
            view_names = inspected.get_view_names(schema=schema_name)
        except NotImplementedError:
            # Some DB providers do not understand 'views'
            self._warn_no_view_detection()
            view_names = []

        objects = [(t, False) for t in table_names] + [(v, True) for v in view_names]
        if self.config.get("lower_case_table_names", True):
            objects = [x.lower() for x in objects]

        return objects

    def to_sql_type(self, jsonschema_type: dict) -> sqlalchemy.types.TypeEngine:  # noqa
        """Convert JSON Schema type to a SQL type.
        Args:
            jsonschema_type: The JSON Schema object.
        Returns:
            The SQL type.
        """
        if self._jsonschema_type_check(jsonschema_type, ("string",)):
            datelike_type = get_datelike_property_type(jsonschema_type)
            if datelike_type:
                if datelike_type == "date-time":
                    return cast(sqlalchemy.types.TypeEngine, mysql.DATETIME())
                elif datelike_type in "time":
                    return cast(sqlalchemy.types.TypeEngine, mysql.TIME())
                elif datelike_type == "date":
                    return cast(sqlalchemy.types.TypeEngine, mysql.DATE())
                elif datelike_type == "binary":
                    return cast(sqlalchemy.types.TypeEngine, mysql.BINARY())

            # The maximum row size for the used table type, not counting BLOBs, is 65535.
            maxlength = jsonschema_type.get("maxLength", 1000)
            data_type = mysql.VARCHAR(maxlength)
            if maxlength <= 1000:
                return cast(sqlalchemy.types.TypeEngine, mysql.VARCHAR(maxlength))
            elif maxlength <= 65535:
                return cast(sqlalchemy.types.TypeEngine, mysql.TEXT(maxlength))
            elif maxlength <= 16777215:
                return cast(sqlalchemy.types.TypeEngine, mysql.MEDIUMTEXT())
            elif maxlength <= 4294967295:
                return cast(sqlalchemy.types.TypeEngine, mysql.LONGTEXT())

            return cast(sqlalchemy.types.TypeEngine, data_type)

        if self._jsonschema_type_check(jsonschema_type, ("integer",)):
            minimum = jsonschema_type.get("minimum", -9223372036854775807)
            maximum = jsonschema_type.get("maximum", 9223372036854775807)

            if minimum >= -128 and maximum <= 127:
                return cast(sqlalchemy.types.TypeEngine, mysql.TINYINT(unsigned=False))
            elif minimum >= -32768 and maximum <= 32767:
                return cast(sqlalchemy.types.TypeEngine, mysql.SMALLINT(unsigned=False))
            elif minimum >= -8388608 and maximum <= 8388607:
                return cast(
                    sqlalchemy.types.TypeEngine, mysql.MEDIUMINT(unsigned=False)
                )
            elif minimum >= -2147483648 and maximum <= 2147483647:
                return cast(sqlalchemy.types.TypeEngine, mysql.INTEGER(unsigned=False))
            elif minimum >= -9223372036854775808 and maximum <= 9223372036854775807:
                return cast(sqlalchemy.types.TypeEngine, mysql.BIGINT(unsigned=False))
            elif minimum >= 0 and maximum <= 255:
                return cast(sqlalchemy.types.TypeEngine, mysql.TINYINT(unsigned=True))
            elif minimum >= 0 and maximum <= 65535:
                return cast(sqlalchemy.types.TypeEngine, mysql.SMALLINT(unsigned=True))
            elif minimum >= 0 and maximum <= 16777215:
                return cast(sqlalchemy.types.TypeEngine, mysql.MEDIUMINT(unsigned=True))
            elif minimum >= 0 and maximum <= 4294967295:
                return cast(sqlalchemy.types.TypeEngine, mysql.INTEGER(unsigned=True))
            elif minimum >= 0 and maximum <= 18446744073709551615:
                return cast(sqlalchemy.types.TypeEngine, mysql.BIGINT(unsigned=True))

        if self._jsonschema_type_check(jsonschema_type, ("number",)):
            if "multipleOf" in jsonschema_type:
                return cast(sqlalchemy.types.TypeEngine, mysql.DECIMAL())
            else:
                return cast(sqlalchemy.types.TypeEngine, mysql.FLOAT())

        if self._jsonschema_type_check(jsonschema_type, ("boolean",)):
            return cast(sqlalchemy.types.TypeEngine, mysql.BOOLEAN())

        if self._jsonschema_type_check(jsonschema_type, ("object",)):
            # if 'format' in jsonschema_type and jsonschema_type.get("format") == "spatial":
            #     return cast(sqlalchemy.types.TypeEngine, mysql.MU)
            return cast(sqlalchemy.types.TypeEngine, mysql.JSON())

        if self._jsonschema_type_check(jsonschema_type, ("array",)):
            return cast(sqlalchemy.types.TypeEngine, sqlalchemy.types.TEXT(4000))

        return cast(sqlalchemy.types.TypeEngine, sqlalchemy.types.TEXT(4000))

    def _jsonschema_type_check(
        self, jsonschema_type: dict, type_check: tuple[str]
    ) -> bool:
        """Return True if the jsonschema_type supports the provided type.
        Args:
            jsonschema_type: The type dict.
            type_check: A tuple of type strings to look for.
        Returns:
            True if the schema suports the type.
        """
        if "type" in jsonschema_type:
            if isinstance(jsonschema_type["type"], (list, tuple)):
                for t in jsonschema_type["type"]:
                    if t in type_check:
                        return True
            else:
                if jsonschema_type.get("type") in type_check:
                    return True

        if any(t in type_check for t in jsonschema_type.get("anyOf", ())):
            return True

        return False

    def _create_empty_column(
        self,
        full_table_name: str,
        column_name: str,
        sql_type: sqlalchemy.types.TypeEngine,
    ) -> None:
        """Create a new column.
        Args:
            full_table_name: The target table name.
            column_name: The name of the new column.
            sql_type: SQLAlchemy type engine to be used in creating the new column.
        Raises:
            NotImplementedError: if adding columns is not supported.
        """
        if not self.allow_column_add:
            raise NotImplementedError("Adding columns is not supported.")

        if column_name.startswith("_"):
            column_name = f"x{column_name}"

        create_column_clause = sqlalchemy.schema.CreateColumn(
            sqlalchemy.Column(
                column_name,
                sql_type,
            )
        )

        try:
            alter_sql = f"""ALTER TABLE {str(full_table_name)}
                ADD COLUMN {str(create_column_clause)} """
            self.logger.info("Altering with SQL: %s", alter_sql)
            self.connection.execute(alter_sql)
        except Exception as e:
            raise RuntimeError(
                f"Could not create column '{create_column_clause}' "
                f"on table '{full_table_name}'."
            ) from e

    def create_temp_table_from_table(self, from_table_name, temp_table_name):
        """Temp table from another table."""

        try:
            self.connection.execute(f"""DROP TABLE {temp_table_name}""")
        except Exception as e:
            pass

        ddl = f"""
            CREATE TABLE {temp_table_name} AS (
                SELECT * FROM {from_table_name}
                WHERE 1=0
            )
        """

        self.connection.execute(ddl)

    def create_empty_table(
        self,
        full_table_name: str,
        schema: dict,
        primary_keys: list[str] | None = None,
        partition_keys: list[str] | None = None,
        as_temp_table: bool = False,
    ) -> None:
        """Create an empty target table.
        Args:
            full_table_name: the target table name.
            schema: the JSON schema for the new table.
            primary_keys: list of key properties.
            partition_keys: list of partition keys.
            as_temp_table: True to create a temp table.
        Raises:
            NotImplementedError: if temp tables are unsupported and as_temp_table=True.
            RuntimeError: if a variant schema is passed with no properties defined.
        """
        if as_temp_table:
            raise NotImplementedError("Temporary tables are not supported.")

        _ = partition_keys  # Not supported in generic implementation.

        _, schema_name, table_name = self.parse_full_table_name(full_table_name)
        meta = sqlalchemy.MetaData(schema=schema_name)
        columns: list[sqlalchemy.Column] = []
        primary_keys = primary_keys or []
        try:
            properties: dict = schema["properties"]
        except KeyError:
            raise RuntimeError(
                f"Schema for '{full_table_name}' does not define properties: {schema}"
            )

        for property_name, property_jsonschema in properties.items():
            is_primary_key = property_name in primary_keys
            columns.append(
                sqlalchemy.Column(property_name, self.to_sql_type(property_jsonschema))
            )

        if primary_keys:
            pk_constraint = PrimaryKeyConstraint(*primary_keys, name=f"{table_name}_PK")
            _ = sqlalchemy.Table(table_name, meta, *columns, pk_constraint)
        else:
            _ = sqlalchemy.Table(table_name, meta, *columns)

        meta.create_all(self._engine)

    def merge_sql_types(  # noqa
        self, sql_types: list[sqlalchemy.types.TypeEngine]
    ) -> sqlalchemy.types.TypeEngine:  # noqa
        """Return a compatible SQL type for the selected type list.
        Args:
            sql_types: List of SQL types.
        Returns:
            A SQL type that is compatible with the input types.
        Raises:
            ValueError: If sql_types argument has zero members.
        """
        if not sql_types:
            raise ValueError("Expected at least one member in `sql_types` argument.")

        if len(sql_types) == 1:
            return sql_types[0]

        # Gathering Type to match variables
        # sent in _adapt_column_type
        current_type = sql_types[0]
        # sql_type = sql_types[1]

        # Getting the length of each type
        # current_type_len: int = getattr(sql_types[0], "length", 0)
        sql_type_len: int = getattr(sql_types[1], "length", 0)
        if sql_type_len is None:
            sql_type_len = 0

        # Convert the two types given into a sorted list
        # containing the best conversion classes
        sql_types = self._sort_types(sql_types)

        # If greater than two evaluate the first pair then on down the line
        if len(sql_types) > 2:
            return self.merge_sql_types(
                [self.merge_sql_types([sql_types[0], sql_types[1]])] + sql_types[2:]
            )

        assert len(sql_types) == 2
        # Get the generic type class
        for opt in sql_types:
            # Get the length
            opt_len: int = getattr(opt, "length", 0)
            generic_type = type(opt.as_generic())

            current_type_length = 0
            if (
                isinstance(current_type, sqlalchemy.types.TEXT)
                and current_type.length is None
            ):
                current_type_length = 65535
            elif hasattr(current_type, "length"):
                current_type_length = current_type.length

            if isinstance(generic_type, type):
                if issubclass(
                    generic_type,
                    (sqlalchemy.types.String, sqlalchemy.types.Unicode),
                ):
                    # If length None or 0 then is varchar max ?
                    if (
                        (opt_len is None)
                        or (opt_len == 0)
                        or (opt_len >= current_type_length)
                    ):
                        return opt
                elif isinstance(
                    generic_type,
                    (sqlalchemy.types.String, sqlalchemy.types.Unicode),
                ):
                    # If length None or 0 then is varchar max ?
                    if (
                        (opt_len is None)
                        or (current_type is None)
                        or (opt_len == 0)
                        or (opt_len >= current_type_length)
                    ):
                        return opt
                # If best conversion class is equal to current type
                # return the best conversion class
                elif str(opt) == str(current_type):
                    return opt

        raise ValueError(
            f"Unable to merge sql types: {', '.join([str(t) for t in sql_types])}"
        )

    def _adapt_column_type(
        self,
        full_table_name: str,
        column_name: str,
        sql_type: sqlalchemy.types.TypeEngine,
    ) -> None:
        """Adapt table column type to support the new JSON schema type.
        Args:
            full_table_name: The target table name.
            column_name: The target column name.
            sql_type: The new SQLAlchemy type.
        Raises:
            NotImplementedError: if altering columns is not supported.
        """
        current_type: sqlalchemy.types.TypeEngine = self._get_column_type(
            full_table_name, column_name
        )

        # Check if the existing column type and the sql type are the same
        if str(sql_type) == str(current_type):
            # The current column and sql type are the same
            # Nothing to do
            return

        # Not the same type, generic type or compatible types
        # calling merge_sql_types for assistnace
        compatible_sql_type = self.merge_sql_types([current_type, sql_type])

        if str(compatible_sql_type).split(" ")[0] == str(current_type).split(" ")[0]:
            # Nothing to do
            return

        if not self.allow_column_alter:
            raise NotImplementedError(
                "Altering columns is not supported. "
                f"Could not convert column '{full_table_name}.{column_name}' "
                f"from '{current_type}' to '{compatible_sql_type}'."
            )
        try:
            alter_sql = f"""ALTER TABLE {str(full_table_name)}
                MODIFY {str(column_name)} {str(compatible_sql_type)}"""
            self.logger.info("Altering with SQL: %s", alter_sql)
            self.connection.execute(alter_sql)
        except Exception as e:
            raise RuntimeError(
                f"Could not convert column '{full_table_name}.{column_name}' "
                f"from '{current_type}' to '{compatible_sql_type}'."
            ) from e


class MySQLSink(SQLSink):
    """MySQL target sink class."""

    connector_class = MySQLConnector

    soft_delete_column_name = "x_sdc_deleted_at"
    version_column_name = "x_sdc_table_version"

    # @property
    # def schema_name(self) -> Optional[str]:
    #     """Return the schema name or `None` if using names with no schema part.
    #     Returns:
    #         The target schema name.
    #     """
    #     # Look for a default_target_scheme in the configuraion fle
    #     default_target_schema: str = self.config.get("default_target_schema", None)
    #     parts = self.stream_name.split("-")

    #     # 1) When default_target_scheme is in the configuration use it
    #     # 2) if the streams are in <schema>-<table> format use the
    #     #    stream <schema>
    #     # 3) Return None if you don't find anything
    #     if default_target_schema:
    #         return default_target_schema

    #     # Schema name not detected.
    #     return None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger.setLevel(logging.DEBUG)

    @property
    def max_size(self) -> int:
        return self.config.get("batch_flush_size")

    def process_record(self, record: dict, context: dict) -> None:
        # Record extracted time for metric calculations
        if "time_extracted" not in record:
            record["time_extracted"] = datetime.now(timezone.utc)
        else:
            record["time_extracted"] = ensure_datetime(record["time_extracted"])
        super().process_record(record, context)

    def emit_state(self, state):
        """Emit state message to standard output then it can be
        consumed by other components"""
        if state is not None:
            line = json.dumps(state)
            self.logger.debug("Emitting state %s", line)
            sys.stdout.write("{}\n".format(line))
            sys.stdout.flush()

    def mark_drained(self) -> None:
        super().mark_drained()
        self.emit_state(self.drained_state)

    def process_batch(self, context: dict) -> None:
        """Process a batch with the given batch context.
        Writes a batch to the SQL target. Developers may override this method
        in order to provide a more efficient upload/upsert process.
        Args:
            context: Stream partition or context dictionary.
        """
        # First we need to be sure the main table is already created
        try:
            conformed_records = (
                [self.conform_record(record) for record in context["records"]]
                if isinstance(context["records"], list)
                else (self.conform_record(record) for record in context["records"])
            )

            join_keys = [
                self.conform_name(key, "column") for key in self.key_properties
            ]
            schema = self.conform_schema(self.schema)

            if self.key_properties:
                self.logger.info(f"Preparing table {self.full_table_name}")
                self.connector.prepare_table(
                    full_table_name=self.full_table_name,
                    schema=schema,
                    primary_keys=self.key_properties,
                    as_temp_table=False,
                )

                tmp_table_name = self.full_table_name + "_temp"

                # Create a temp table (Creates from the table above)
                self.logger.info(f"Creating temp table {self.full_table_name}")
                self._connector.create_temp_table_from_table(
                    from_table_name=self.full_table_name, temp_table_name=tmp_table_name
                )

                # Insert into temp table
                self.bulk_insert_records(
                    full_table_name=tmp_table_name,
                    schema=schema,
                    records=conformed_records,
                )
                # Merge data from Temp table to main table
                self.logger.info(
                    f"Merging data from temp table to {self.full_table_name}"
                )
                self.merge_upsert_from_table(
                    from_table_name=tmp_table_name,
                    to_table_name=self.full_table_name,
                    join_keys=join_keys,
                )

            else:
                self.bulk_insert_records(
                    full_table_name=self.full_table_name,
                    schema=schema,
                    records=conformed_records,
                )
            # calculate metrics
            export_time = datetime.now(timezone.utc)
            for r in conformed_records:
                lag = export_time - r["time_extracted"]
                export_lag.labels(
                    region_label, tenant_label, fabric_label, workflow_label
                ).set(lag.total_seconds())
        except Exception as e:
            # Increment export_errors metric
            export_errors.labels(
                region_label, tenant_label, fabric_label, workflow_label
            ).inc()
            raise e

    def merge_upsert_from_table(
        self,
        from_table_name: str,
        to_table_name: str,
        join_keys: List[str],
    ) -> Optional[int]:
        """Merge upsert data from one table to another.
        Args:
            from_table_name: The source table name.
            to_table_name: The destination table name.
            join_keys: The merge upsert keys, or `None` to append.
            schema: Singer Schema message.
        Return:
            The number of records copied, if detectable, or `None` if the API does not
            report number of records affected/inserted.
        """
        # TODO think about sql injeciton,
        # issue here https://github.com/MeltanoLabs/target-postgres/issues/22

        join_keys = [self.conform_name(key, "column") for key in join_keys]
        schema = self.conform_schema(self.schema)

        join_condition = " and ".join(
            [f"temp.{key} = target.{key}" for key in join_keys]
        )

        upsert_on_condition = " and ".join(
            [f"{key} = VALUES({key})" for key in join_keys]
        )

        merge_sql = f"""
            INSERT INTO {to_table_name} ({", ".join(schema["properties"].keys())})
                SELECT {", ".join(schema["properties"].keys())}
                FROM 
                    {from_table_name} temp
            ON DUPLICATE KEY UPDATE 
                {upsert_on_condition}
        """

        self.logger.info("Merging with SQL: %s", merge_sql)

        self.connection.execute(merge_sql)

        self.connection.execute("COMMIT")

        self.connection.execute(f"DROP TABLE {from_table_name}")

    def bulk_insert_records(
        self,
        full_table_name: str,
        schema: dict,
        records: Iterable[Dict[str, Any]],
    ) -> Optional[int]:
        """Bulk insert records to an existing destination table.
        The default implementation uses a generic SQLAlchemy bulk insert operation.
        This method may optionally be overridden by developers in order to provide
        faster, native bulk uploads.
        Args:
            full_table_name: the target table name.
            schema: the JSON schema for the new table, to be used when inferring column
                names.
            records: the input records.
        Returns:
            True if table exists, False if not, None if unsure or undetectable.
        """
        insert_sql = self.generate_insert_statement(
            full_table_name,
            schema,
        )
        if isinstance(insert_sql, str):
            insert_sql = sqlalchemy.text(insert_sql)

        self.logger.info("Inserting with SQL: %s", insert_sql)

        columns = self.column_representation(schema)

        # temporary fix to ensure missing properties are added
        insert_records = []

        for record in records:
            insert_record = {}
            conformed_record = self.conform_record(record)
            for column in columns:
                # insert_record[column.name] = conformed_record.get(column.name)

                val = conformed_record.get(column.name)
                if isinstance(val, Dict) or isinstance(val, List):
                    val = json.dumps(val)

                insert_record[column.name] = val
            insert_records.append(insert_record)

        self.connection.execute(insert_sql, insert_records)
        self.connection.execute("COMMIT")

        if isinstance(records, list):
            return len(records)  # If list, we can quickly return record count.

        return None  # Unknown record count.

    def column_representation(
        self,
        schema: dict,
    ) -> List[Column]:
        """Returns a sql alchemy table representation for the current schema."""
        columns: list[Column] = []
        conformed_properties = self.conform_schema(schema)["properties"]
        for property_name, property_jsonschema in conformed_properties.items():
            columns.append(
                Column(
                    property_name,
                    self.connector.to_sql_type(property_jsonschema),
                )
            )
        return columns

    def snakecase(self, name):
        name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
        name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", name)
        return name.lower()

    def move_leading_underscores(self, text):
        match = re.match(r"^(_*)(.*)", text)
        if match:
            result = match.group(2) + match.group(1)
            return result
        return text

    def conform_name(self, name: str, object_type: Optional[str] = None) -> str:
        """Conform a stream property name to one suitable for the target system.
        Transforms names to snake case by default, applicable to most common DBMSs'.
        Developers may override this method to apply custom transformations
        to database/schema/table/column names.
        Args:
            name: Property name.
            object_type: One of ``database``, ``schema``, ``table`` or ``column``.
        Returns:
            The name transformed to snake case.
        """
        # strip non-alphanumeric characters except _.
        name = re.sub(r"[^a-zA-Z0-9_]+", "_", name)

        # Move leading underscores to the end of the name
        name = self.move_leading_underscores(name)

        # convert to snakecase
        name = self.snakecase(name)
        # replace leading digit
        return replace_leading_digit(name)



import pkg_resources

from c8connector import C8Connector, ConfigProperty, ConfigAttributeType, Sample, Schema


class MySQLTargetConnector(C8Connector):
    """MySQLTargetConnector's C8Connector impl."""

    def name(self) -> str:
        """Returns the name of the connector."""
        return "MySQL"

    def package_name(self) -> str:
        """Returns the package name of the connector (i.e. PyPi package name)."""
        return "macrometa-target-mysql"

    def version(self) -> str:
        """Returns the version of the connector."""
        return pkg_resources.get_distribution("macrometa_target_mysql").version

    def type(self) -> str:
        """Returns the type of the connector."""
        return "target"

    def description(self) -> str:
        """Returns the description of the connector."""
        return "Send data into a MySQL table."

    def logo(self) -> str:
        """Returns the logo image for the connector."""
        return ""

    def validate(self, integration: dict) -> None:
        """Validate given configurations against the connector.
        If invalid, throw an exception with the cause.
        """
        pass

    def samples(self, integration: dict) -> list[Sample]:
        """Fetch sample data using the given configurations."""
        return []

    def schemas(self, integration: dict) -> list[Schema]:
        """Get supported schemas using the given configurations."""
        return []

    def reserved_keys(self) -> list[str]:
        """List of reserved keys for the connector."""
        return []

    def config(self) -> list[ConfigProperty]:
        """Get configuration parameters for the connector."""
        return [
            ConfigProperty(
                "host",
                "Host",
                ConfigAttributeType.STRING,
                True,
                False,
                description="MySQL host.",
                placeholder_value="mysql_host",
            ),
            ConfigProperty(
                "port",
                "Port",
                ConfigAttributeType.INT,
                True,
                False,
                description="MySQL port.",
                default_value="3306",
            ),
            ConfigProperty(
                "username",
                "Username",
                ConfigAttributeType.STRING,
                True,
                False,
                description="MySQL user.",
                placeholder_value="mysql",
            ),
            ConfigProperty(
                "password",
                "Password",
                ConfigAttributeType.PASSWORD,
                True,
                False,
                description="MySQL password.",
                placeholder_value="password",
            ),
            ConfigProperty(
                "database",
                "Database Name",
                ConfigAttributeType.STRING,
                True,
                True,
                description="MySQL database name.",
                placeholder_value="mysql_db",
            ),
            ConfigProperty(
                "target_table",
                "Target Table",
                ConfigAttributeType.STRING,
                True,
                True,
                description="Destination table name.",
                placeholder_value="mysql_table",
            ),
            ConfigProperty(
                "replace_null",
                "Replace Null",
                ConfigAttributeType.BOOLEAN,
                False,
                False,
                description="Replace null values with default values.",
                default_value="false",
            ),
            ConfigProperty(
                "batch_flush_size",
                "Batch Size",
                ConfigAttributeType.INT,
                False,
                False,
                description="Maximum size of batch. Exceeding this will trigger a batch flush.",
                default_value="10000",
            ),
            ConfigProperty(
                "batch_flush_interval",
                "Batch Flush Interval (Minutes)",
                ConfigAttributeType.INT,
                False,
                False,
                description="Time between batch flush executions.",
                default_value="5",
            ),
            ConfigProperty(
                "lower_case_table_names",
                "Lower Case Tables",
                ConfigAttributeType.BOOLEAN,
                False,
                True,
                description="Use lowercase for table names or not. This will target table name "
                "to lower case",
                default_value="false",
            ),
            ConfigProperty(
                "allow_column_alter",
                "Allow Column Alter",
                ConfigAttributeType.BOOLEAN,
                False,
                True,
                description="Allow column alterations if target table's column type doesn't match "
                "the source data type.",
                default_value="false",
            ),
        ]

    def capabilities(self) -> list[str]:
        """Return the capabilities[1] of the connector."""
        return []



""" Attempt at making some standard Target Tests. """
import io
import json

# flake8: noqa
import os
from contextlib import redirect_stdout
from pathlib import Path

import pytest
import sqlalchemy
from jsonschema.exceptions import ValidationError
from singer_sdk.exceptions import (
    MissingKeyPropertiesError,
    RecordsWithoutSchemaException,
)
from sqlalchemy import create_engine

from macrometa_target_mysql.target import MacrometaTargetMySQL

parent_directory = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "../../tests/")
)
config_path = os.path.join(parent_directory, "config.json")

with open(config_path, "r") as f:
    config_data: dict = json.load(f)


@pytest.fixture()
def mysql_config():
    return config_data


@pytest.fixture
def mysql_target(mysql_config) -> MacrometaTargetMySQL:
    return MacrometaTargetMySQL(config=mysql_config)


def singer_file_to_target(file_name, target) -> None:
    """Singer file to Target, emulates a tap run
    Equivalent to running cat file_path | target-name --config config.json.
    Note that this function loads all lines into memory, so it is
    not good very large files.
    Args:
        file_name: name to file in .tests/data_files to be sent into target
        target: Target to pass data from file_path into..
    """
    file_path = Path(__file__).parent / Path("./data_files") / Path(file_name)
    buf = io.StringIO()
    with redirect_stdout(buf):
        with open(file_path, "r") as f:
            for line in f:
                print(line.rstrip("\r\n"))  # File endings are here,
                # and print adds another line ending so we need to remove one.
    buf.seek(0)
    target.listen(buf)


def get_engine():
    connection_url = sqlalchemy.engine.url.URL.create(
        drivername="mysql",
        username=config_data["user"],
        password=config_data["password"],
        host=config_data["host"],
        port=config_data["port"],
        database=config_data["database"],
    )

    engine = create_engine(connection_url)

    return engine


def get_row_count(table_name):
    engine = get_engine()
    rowcount = engine.execute(f"SELECT COUNT(*) FROM {table_name}").fetchone()[0]
    return rowcount


def get_table_cols(table_name):
    engine = get_engine()
    q = f"""
    SELECT COLUMN_NAME
    FROM INFORMATION_SCHEMA.COLUMNS
    WHERE TABLE_SCHEMA='{config_data["database"]}'
    AND TABLE_NAME='{table_name}'
    ORDER BY ORDINAL_POSITION;
    """
    columns = [col[0] for col in engine.execute(q).fetchall()]
    return columns


def drop_table(table_name):
    engine = get_engine()
    engine.execute(f"DROP TABLE if exists {table_name}")


# this test should throw an exception
def test_record_before_schema(mysql_target):
    with pytest.raises(RecordsWithoutSchemaException) as e_info:
        file_name = "record_before_schema.singer"
        singer_file_to_target(file_name, mysql_target)


# this test should throw an exception
def test_invalid_schema(mysql_target):
    with pytest.raises(Exception) as e_info:
        file_name = "invalid_schema.singer"
        singer_file_to_target(file_name, mysql_target)


# this test should throw an exception
def test_record_missing_key_property(mysql_target):
    with pytest.raises(Exception) as e_info:
        file_name = "record_missing_key_property.singer"
        singer_file_to_target(file_name, mysql_target)


# this test should throw an exception
def test_record_missing_required_property(mysql_target):
    with pytest.raises(Exception) as e_info:
        file_name = "record_missing_required_property.singer"
        singer_file_to_target(file_name, mysql_target)


# test that data is correctly set
# @pytest.mark.skip(reason="Waiting for SDK to handle this")
def test_special_chars_in_attributes(mysql_target):
    with pytest.raises(MissingKeyPropertiesError) as e_info:
        file_name = "special_chars_in_attributes.singer"
        singer_file_to_target(file_name, mysql_target)


# test that data is correctly set
def test_optional_attributes(mysql_target):
    file_name = "optional_attributes.singer"
    singer_file_to_target(file_name, mysql_target)


def test_schema_no_properties(mysql_target):
    file_name = "schema_no_properties.singer"
    singer_file_to_target(file_name, mysql_target)


# test that data is correct
def test_schema_updates_1(mysql_target):
    file_name = "schema_updates.singer"

    drop_table("test_schema_updates")

    conf_key = "allow_column_alter"
    orig_conf = config_data.get(conf_key, True)

    config_data[conf_key] = False
    mysql_target = MacrometaTargetMySQL(config=config_data)
    with pytest.raises(NotImplementedError) as e_info:
        singer_file_to_target(file_name, mysql_target)

    config_data[conf_key] = orig_conf


def test_schema_updates_2(mysql_target):
    file_name = "schema_updates.singer"

    drop_table("test_schema_updates")

    conf_key = "allow_column_alter"
    orig_conf = config_data.get(conf_key, True)

    config_data[conf_key] = True
    mysql_target = MacrometaTargetMySQL(config=config_data)
    singer_file_to_target(file_name, mysql_target)

    config_data[conf_key] = orig_conf


# test that data is correct
def test_multiple_state_messages(mysql_target):
    file_name = "multiple_state_messages.singer"
    singer_file_to_target(file_name, mysql_target)


def test_relational_data(mysql_target):
    file_name = "user_location_data.singer"
    singer_file_to_target(file_name, mysql_target)

    file_name = "user_location_upsert_data.singer"
    singer_file_to_target(file_name, mysql_target)


# test that data is correct
def test_no_primary_keys(mysql_target):
    file_name = "no_primary_keys.singer"
    singer_file_to_target(file_name, mysql_target)

    file_name = "no_primary_keys_append.singer"
    singer_file_to_target(file_name, mysql_target)


# test that data is correct
def test_duplicate_records(mysql_target):
    file_name = "duplicate_records.singer"
    singer_file_to_target(file_name, mysql_target)


# test that data is correct
def test_update_records(mysql_target):
    file_name = "update_records.singer"
    singer_file_to_target(file_name, mysql_target)


def test_array_data(mysql_target):
    file_name = "array_data.singer"
    singer_file_to_target(file_name, mysql_target)


def test_encoded_string_data(mysql_target):
    file_name = "encoded_strings.singer"
    singer_file_to_target(file_name, mysql_target)


@pytest.mark.skip(reason="Something about objects not supported")
def test_tap_appl(mysql_target):
    file_name = "tap_aapl.singer"
    singer_file_to_target(file_name, mysql_target)


def test_missing_value(mysql_target):
    file_name = "missing_value.singer"
    singer_file_to_target(file_name, mysql_target)


def test_large_int(mysql_target):
    file_name = "large_int.singer"
    singer_file_to_target(file_name, mysql_target)


def test_db_schema(mysql_target):
    file_name = "target_schema.singer"
    singer_file_to_target(file_name, mysql_target)


def test_illegal_colnames(mysql_target):
    with pytest.raises(MissingKeyPropertiesError) as e_info:
        file_name = "illegal_colnames.singer"
        singer_file_to_target(file_name, mysql_target)


def test_numerics(mysql_target):
    file_name = "numerics.singer"
    singer_file_to_target(file_name, mysql_target)


def test_integer_null_value(mysql_target):
    file_name = "integer_null_value.singer"

    orig_conf = config_data["replace_null"]

    config_data["replace_null"] = True
    singer_file_to_target(file_name, mysql_target)

    with pytest.raises(ValidationError) as e_info:
        config_data["replace_null"] = False
        singer_file_to_target(file_name, mysql_target)

    config_data["replace_null"] = orig_conf


def test_string_null_value(mysql_target):
    file_name = "string_null_value.singer"

    orig_conf = config_data["replace_null"]

    config_data["replace_null"] = True
    singer_file_to_target(file_name, mysql_target)

    with pytest.raises(ValidationError) as e_info:
        config_data["replace_null"] = False
        singer_file_to_target(file_name, mysql_target)

    config_data["replace_null"] = orig_conf



import os

from prometheus_client import CollectorRegistry, Counter, Gauge, start_http_server

# Create a custom CollectorRegistry
registry_package = CollectorRegistry()
export_lag = Gauge(
    "export_lag",
    "The average time from when the data changes in GDN collections are reflected in external data sources",
    ["region", "tenant", "fabric", "workflow"],
    registry=registry_package,
)
export_errors = Counter(
    "export_errors",
    "Total count of errors while exporting data from GDN collections",
    ["region", "tenant", "fabric", "workflow"],
    registry=registry_package,
)

region_label = os.getenv("GDN_FEDERATION", "NA")
tenant_label = os.getenv("GDN_TENANT", "NA")
fabric_label = os.getenv("GDN_FABRIC", "NA")
workflow_label = os.getenv("WORKFLOW_UUID", "NA")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['macrometa_target_mysql', 'macrometa_target_mysql.tests']

package_data = \
{'': ['*'], 'macrometa_target_mysql.tests': ['data_files/*']}

install_requires = \
['c8connector==0.0.32',
 'cryptography>=41.0.2,<42.0.0',
 'prometheus-client==0.16.0',
 'pymysql>=1.1.0,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'singer-sdk>=0.30.0,<0.31.0']

entry_points = \
{'console_scripts': ['macrometa-target-mysql = '
                     'macrometa_target_mysql.target:MacrometaTargetMySQL.cli']}

setup_kwargs = {
    'name': 'macrometa-target-mysql',
    'version': '0.0.4',
    'description': 'A Meltano target for MySQL.',
    'long_description': '# target-mysql\n\n`target-mysql` is a Singer target for Oracle, Build with the [Meltano Target SDK](https://sdk.meltano.com).\n\n\n\nEnglish | [한국어](./docs/README_ko.md)\n\n\n## Installation\n\nUse PIP for installation:\n\n```bash\npip install thk-target-mysql\n```\n\nOr use GitHub Repo:\n\n```bash\npipx install git+https://github.com/thkwag/target-mysql.git@main\n```\n\n## Configuration\n\nThe available configuration options for `target-mysql` are:\n\n| Configuration Options   | Description                                | Default            |\n|-------------------------|--------------------------------------------|--------------------|\n| host                    | MySQL server\'s hostname or IP address      |                    |\n| port                    | Port where MySQL server is running         |                    |\n| user                    | MySQL username                             |                    |\n| password                | MySQL user\'s password                      |                    |\n| database                | MySQL database\'s name                      |                    |\n| table_name_pattern      | MySQL table name pattern                   | "${TABLE_NAME}"    |\n| lower_case_table_names  | Use lowercase for table names or not       | true               |\n| allow_column_alter      | Allow column alterations or not            | false              |\n| replace_null            | Replace null values with others or not     | false              |\n\nConfigurations can be stored in a JSON configuration file and specified using the `--config` flag with `target-mysql`.\n\n### The `replace_null` Option (Experimental)\n\nBy enabling the `replace_null` option, null values are replaced with \'empty\' equivalents based on their data type. Use with caution as it may alter data semantics.\n\nWhen `replace_null` is `true`, null values are replaced as follows:\n\n| JSON Schema Data Type | Null Value Replacement |\n|-----------------------|------------------------|\n| string                | Empty string(`""`)     |\n| number                | `0`                    |\n| object                | Empty object(`{}`)     |\n| array                 | Empty array(`[]`)      |\n| boolean               | `false`                |\n| null                  | null                   |\n\n\n## Usage\n\n```bash\ncat <input_stream> | target-mysql --config <config.json>\n```\n\n- `<input_stream>`: Input data stream\n- `<config.json>`: JSON configuration file\n\n`target-mysql` reads data from a Singer Tap and writes it to a MySQL database. Run Singer Tap to generate data before launching `target-mysql`.\n\nHere\'s an example of using Singer Tap with `target-mysql`:\n\n```bash\ntap-exchangeratesapi | target-mysql --config config.json\n```\n\nIn this case, `tap-exchangeratesapi` is a Singer Tap that generates exchange rate data. The data is passed to `target-mysql` through a pipe(`|`), and `target-mysql` writes it to a MySQL database. `config.json` contains `target-mysql` settings.\n\n## Developer Resources\n\n### Initializing the Development Environment\n\n```bash\npipx install poetry\npoetry install\n```\n\n### Creating and Running Tests\n\nCreate tests in the `macrometa_target_mysql/tests` subfolder and run:\n\n```bash\npoetry run pytest\n```\n\nUse `poetry run` to test `target-mysql` CLI interface:\n\n```bash\npoetry run target-mysql --help\n```\n\n### Testing with [Meltano](https://meltano.com/)\n\n_**Note:** This target functions within a Singer environment and does not require Meltano._\n\nFirstly, install Meltano and necessary plugins:\n\n```bash\n# Install Meltano\npipx install meltano\n\n# Initialize Meltano in this directory\ncd target-mysql\nmeltano install\n```\n\nThen, test and orchestrate with Meltano:\n\n```bash\n# Call tests:\nmeltano invoke target-mysql --version\n\n# Or execute pipeline with Carbon Intensity sample tap:\nmeltano run tap-carbon-intensity target-mysql\n```\n\n### SDK Development Guide\n\nFor in-depth instructions on crafting Singer Taps and Targets using Meltano Singer SDK, see the [Development Guide](https://sdk.meltano.com/en/latest/dev_guide.html).\n\n## Reference Links\n\n- [Meltano Target SDK Documentation](https://sdk.meltano.com)\n- [Singer Specification](https://github.com/singer-io/getting-started/blob/master/docs/SPEC.md)\n- [Meltano](https://meltano.com/)\n- [Singer.io](https://www.singer.io/)',
    'author': 'Macrometa',
    'author_email': 'info@macrometa.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.1,<3.11',
}


setup(**setup_kwargs)



"""MySQL target class."""

from __future__ import annotations

import io
import typing as t
from pathlib import PurePath

import simplejson as json
from prometheus_client import start_http_server
from singer_sdk import typing as th
from singer_sdk.target_base import SQLTarget

from macrometa_target_mysql.constants import (
    export_errors,
    fabric_label,
    region_label,
    registry_package,
    tenant_label,
    workflow_label,
)
from macrometa_target_mysql.sinks import MySQLSink


class MacrometaTargetMySQL(SQLTarget):
    """Sample target for MySQL."""

    name = "macrometa-target-mysql"

    default_sink_class = MySQLSink

    def __init__(
        self,
        *,
        config: dict | PurePath | str | list[PurePath | str] | None = None,
        parse_env_config: bool = False,
        validate_config: bool = True,
    ) -> None:
        super().__init__(
            config=config,
            parse_env_config=parse_env_config,
            validate_config=validate_config,
        )
        self._MAX_RECORD_AGE_IN_MINUTES = self.config.get("batch_flush_interval")

        # Start the Prometheus HTTP server for exposing metrics
        self.logger.info("MySQL target is starting the metrics server.")
        start_http_server(8001, registry=registry_package)

    config_jsonschema = th.PropertiesList(
        th.Property(
            "username",
            th.StringType,
            secret=True,  # Flag config as protected.
            description="MySQL username",
        ),
        th.Property(
            "password",
            th.StringType,
            secret=True,  # Flag config as protected.
            description="MySQL password",
        ),
        th.Property(
            "host",
            th.StringType,
            description="MySQL host",
        ),
        th.Property(
            "port",
            th.IntegerType,
            description="MySQL port",
        ),
        th.Property(
            "database",
            th.StringType,
            description="MySQL database",
        ),
        th.Property(
            "target_table",
            th.StringType,
            description="MySQL table name",
        ),
        th.Property(
            "lower_case_table_names",
            th.BooleanType,
            description="Lower case table names",
            default=True,
        ),
        th.Property(
            "allow_column_alter",
            th.BooleanType,
            description="Allow column alter",
            default=False,
        ),
        th.Property(
            "replace_null",
            th.BooleanType,
            description="Replace null to blank",
            default=False,
        ),
        th.Property(
            "batch_flush_size", th.IntegerType, description="Batch Size", default=False
        ),
        th.Property(
            "batch_flush_interval",
            th.IntegerType,
            description="Batch Flush Interval (Minutes)",
            default=False,
        ),
        th.Property(
            "batch_flush_size", th.IntegerType, description="Batch Size", default=10000
        ),
        th.Property(
            "batch_flush_interval",
            th.IntegerType,
            description="Batch Flush Interval (Minutes)",
            default=5,
        ),
    ).to_dict()

    schema_properties = {}

    def _process_lines(self, file_input: t.IO[str]) -> t.Counter[str]:
        if self.config.get("replace_null", False):
            processed_input = io.StringIO()
            for line in file_input:
                data = self.deserialize_json(line.strip())

                if data.get("type", "") == "SCHEMA":
                    self.schema_properties = data["schema"]["properties"]
                elif data.get("type", "") == "RECORD":
                    for key, value in data.get("record", {}).items():
                        if value is not None:
                            continue

                        # https://json-schema.org/understanding-json-schema/reference/type.html
                        _type = self.schema_properties[key]["type"]
                        data_types = _type if isinstance(_type, list) else [_type]

                        if "null" in data_types:
                            continue
                        if "string" in data_types:
                            data["record"][key] = ""
                        elif "object" in data_types:
                            data["record"][key] = {}
                        elif "array" in data_types:
                            data["record"][key] = []
                        elif "boolean" in data_types:
                            data["record"][key] = False
                        else:
                            data["record"][key] = 0

                processed_input.write(json.dumps(data) + "\n")
            processed_input.seek(0)
            try:
                return super()._process_lines(processed_input)
            except Exception as e:
                # Increment export_errors metric
                export_errors.labels(
                    region_label, tenant_label, fabric_label, workflow_label
                ).inc()
                raise e
        else:
            try:
                return super()._process_lines(file_input)
            except Exception as e:
                # Increment export_errors metric
                export_errors.labels(
                    region_label, tenant_label, fabric_label, workflow_label
                ).inc()
                raise e


if __name__ == "__main__":
    MacrometaTargetMySQL.cli()
